// This file is part of the CardioVascular toolbox
// Copyright (C) 2012 - INRIA - Serge Steer
// Copyright (C) - INRIA - Alessandro Monti, Claire Médigue, Michel Sorine
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

function [signals_energy,signals_dispersion,inout_gains,inout_coherences]=baroreflex_energy(signals,frequency_bands,options)
//Multichannel analysis
 
//INPUTS
//signals : a N by Nt array with as many rows as signals.  The first
//          row must be the RR signal (output) and the next ones the
//          inputs.
//frequency_bands: a m by 2 array, each row [fmin fmax] specifies the
//          frequency range to be studied
//options : a struct with fields:
  
//     sectionlength: an integer, the signal section length to be used for
//           power and cross spectrum computation.
  
//     sectionstep: an integer, the step to be applied for one signal
//           section to the next one. sectionstep must be less or equal to
//           sectionlength. Overlap is sectionlength-sectionstep
  
//     smoothwindowlength: an integer, the length of the smoothing window
//           (hanning)
  
//     minimalcoherence: a positive number in ]0 1]
  
//signals_energy: a 2D array with N rows. Each row gives
//          the corresponding signal energy
//signals_dispersion: a tlist with fields:
//     fmin: a 2D array with N rows. Each row gives the frequency around
//          which the dispersion is minimal for the corresponding signal.
//     pmin: a 2D array with N rows. Each row gives the value of the SPSD at
//          the frequency of minimal dispersion for the corresponding signal. 
//     dmin: a 2D array with N rows. Each row gives the value defined within
//          the interval [0;1] that describes the dispersion of the
//          spectra. Dispersion value is zero when the signal is merely a
//          sinusoide, only one components, whereas is 1 if the spectra is
//          completely spread, so there isn't a principal component.
//inout_gains : a 2D array with N-1 rows. The input/output transfer
//          function gain
//inout_coherences : a 2D array with N-1 rows. The input/output coherence


//References: Rapport de recherche INRIA RR-4427:"Short term control of
//      the cardivascular system: modelling and signal analysis"
//      Alessandro Monti, Claire Médigue, Michel Sorine,2002 pp 50-
 
  [section_length,section_step,smooth_window_length,minimal_coherence]= CheckBaroreflexOptions(options)
  
  ind=zeros(frequency_bands);
  for l=1:size(frequency_bands,1)
    frequency_band=frequency_bands(l,:);
    ind(l,1)=min(max(ceil(frequency_band(1)*section_length)+1,1),section_length);
    ind(l,2)=min(max(floor(frequency_band(2)*section_length)+1,1),section_length);
  end
  [Nsignals,Npoints]=size(signals);
  //smooth_window_length must be  less  than N and odd (to be ckecked) 
   section_window=window("hm",section_length);
   //section_window=section_window/sum(section_window);
   smoothing_window=window("hn",smooth_window_length);
   smoothing_window=smoothing_window/sum(smoothing_window);
   
   //loop on batchs
   nbatch=floor((Npoints-section_length)/section_step)
   nfb=size(frequency_bands,1)


   //allocate arrays 
   //signal properties
   signals_dispersion_fmin=zeros(Nsignals,nbatch,nfb);
   signals_dispersion_pmin=zeros(Nsignals,nbatch,nfb);
   signals_dispersion_dmin=zeros(Nsignals,nbatch,nfb);
   signals_energy=zeros(Nsignals,nbatch,nfb);
   //inputs->output cross properties
   inout_gains = zeros(Nsignals-1,nbatch,nfb);
   inout_coherences= zeros(Nsignals-1,nbatch,nfb);
   i=0;
   for k=1:section_step:Npoints-section_length
     i=i+1;
     sect=k:k+section_length-1
     for ksig=1:Nsignals
       //Complex spectrum 
       sig_fft=fft(signals(ksig,sect).*section_window)
       //smoothed power spectral density
       spsd=Smooth(abs(sig_fft)^2,smoothing_window)
       if ksig==1 then 
         // preserve output signal data needed for cross properties
         sig_fft_out=sig_fft
         spsd_out=spsd
       end
       for l=1:size(frequency_bands,1)
         //signals dispersion 
         [fmin,pmin,dmin]=DispersionBand(spsd,ind(l,1),ind(l,2));
         signals_dispersion_fmin(ksig,i,l)=fmin;
         signals_dispersion_pmin(ksig,i,l)=pmin;
         signals_dispersion_dmin(ksig,i,l)=dmin;
         //signals energy 
         signals_energy(ksig,i,l)=sum(spsd(ind(l,1):ind(l,2)))/section_length
         if ksig>1 then
           //cross properties
           scsd=Smooth(sig_fft_out.*conj(sig_fft),smoothing_window)
           //input->output gain
           gain=abs(scsd)./spsd
           inout_gains(ksig-1,i,l)=mean(gain(ind(l,1):ind(l,2)));
           //input->output coherence
           coherence=abs(scsd)^2 ./ (spsd_out.*spsd)
           klast=find(coherence(ind(l,1):ind(l,2))<minimal_coherence(l),1);
           if klast==[] then klast=ind(l,2),else klast=ind(l,1)-1+klast;end
           inout_coherences(ksig-1,i,l)=mean(coherence(ind(l,1):klast));
         end
       end
      
     end
   end
   signals_dispersion=tlist(["sigdisp","fmin","pmin","dmin"],...
                            signals_dispersion_fmin/section_length,...
                            signals_dispersion_pmin,...
                            signals_dispersion_dmin)
    
endfunction
//local functions

function xs=Smooth(x,win)
  n=size(win,'*');
  half=(n-1)/2;
  xs=convol(win,x)
  xs=xs(1+half:$-half)
endfunction

function [fmin,pmin,dmin]=DispersionBand(signal,indmin,indmax)
//fmin: frequency  that gives the minimum dispersion
//pmin: associated power
//dmin: associated dispersion

  fmin=0
  pmin=0
  dmin=-1
  
  fc=indmin;//normalized frequency
   for i=indmin:indmax
    pc=signal(i)
    if signal(i)<>0 then
      if indmin<indmax
        f=(i-(indmin:indmax)).^2;//normalized freqency
        dc=sum(f.*signal(indmin:indmax))./sum(f*pc)
      else
        dc=1
      end
    else
      dc=dmin
    end
    if dc<dmin|(dmin==-1&dc<>dmin) then
      fmin=fc
      pmin=pc
      dmin=dc
    end
  end
endfunction




function [section_length,section_step,smooth_window_length,minimal_coherence]=CheckBaroreflexOptions(options)

  err_msg3=_("%s: Wrong type for input argument #%d: A real positive number expected.\n");
  err_msg5=_("%s: Wrong value for field %s of input argument #%d: A power of two expected.\n")
  err_msg6=_("%s: Wrong value for input argument #%d: A struct with fields %s expected.\n")

  
  //default values
  section_length=256
  section_step=128 //50% overlap
  smooth_window_length=5;
  minimal_coherence=0.5;

  if typeof(options)=="st" then
    legal_fields=["sectionlength","sectionstep","smoothwindowlength","minimalcoherence"];
    fn=fieldnames(options)
    for k=1:size(fn,'*')
      select fn(k)
        
      case "sectionlength" then
        //length of batch section
        o=options(fn(k));
        if type(o)<>1|~isreal(o)|size(o,'*')<>1|o<1 then 
          error(msprintf(err_msg3,"baroreflex",fn(k),5))
        end
        // if o<>2^round(log2(o)) then
        //   error(msprintf(err_msg5,"baroreflex",fn(k),5))
        // end
        section_length=o
      case "sectionstep"  then
        //shift from batch to batch
        o=options(fn(k));
        if type(o)<>1|~isreal(o)|size(o,'*')<>1|o<1 then 
          error(msprintf(err_msg3,"baroreflex",fn(k),5))
        end
        section_step=o
      case "smoothwindowlength"  then
        //smoothing window length
        o=options(fn(k));
        if type(o)<>1|~isreal(o)|size(o,'*')<>1|o<1 then 
          error(msprintf(err_msg3,"baroreflex",fn(k),5))
        end
        smooth_window_length=o
      case "minimalcoherence"  then
        //minimal coherence
        o=options(fn(k));
        if type(o)<>1|~isreal(o)|or(o<=0) then 
          error(msprintf(err_msg3,"baroreflex",fn(k),5))
        end
        minimal_coherence=o
      else
        error(msprintf(err_msg6,"baroreflex",5,strcat(legal_fields,",")))
      end
    end
  else
    error(msprintf(err_msg6,"baroreflex",5,strcat(legal_fields,",")))
  end
endfunction
